/*
 * pm_transformed_task.cc
 *
 *  Created on: 13.10.2015
 *      Author: Raphael Imahorn <raphael.imahorn@stud.unibas.ch>
 */

#include "pm_transformed_task.h"

#include "option_parser.h"
#include "plugin.h"
#include "utilities.h"
#include "task_tools.h"
#include "global_operator.h"
#include "globals.h"
#include "timer.h"

#include <algorithm>
#include <iostream>
#include <memory>
#include <list>
#include <utility>
#include <cassert>

using namespace std;

PmTransformedTask::PmTransformedTask(const Options &opts) :
		DelegatingTask(opts.get<shared_ptr<AbstractTask>>("transform")), m(opts.get<int>("m")),
		parentProxy(TaskProxy(*parent)) {
	Timer trafo_timer;
	verify_no_axioms(parentProxy);
	verify_no_conditional_effects(parentProxy);
	parent_num_variables=parent->get_num_variables();
	parent_num_operators=parent->get_num_operators();
	init();
	trafo_timer.stop();
	cout << "Transformation time: " << trafo_timer << endl;
}

void PmTransformedTask::init() {
	for(auto var : parentProxy.get_variables()){
		parent_num_facts += var.get_domain_size();
	}
	if(m <= parent_num_facts){
		cout << "m = " << m << endl;
	}else{
		cout << "m = " << m << " was bigger than the number of facts in the original problem: "
				<< parent_num_facts << ". Therefore m is now: " << parent_num_facts << "." << endl;
		m = parent_num_facts;
	}
	init_atoms();
	init_meta_atoms();
	init_findmap();
	init_operators();
	init_initial_state();
	init_goals();
}

void PmTransformedTask::init_meta_atoms(){
	unsigned temp = 0;
	for(unsigned i = 1; i <= m; ++i){
	    temp += bin(atoms.size(), i);
	}
	// TODO Methode Schreiben, welche suche abbricht,
	// ... falls mehr Meta-Atome entstehen als gespeichert werden können
	cout << "Meta-Atom maximal Zahl: " << meta_atoms.max_size() << endl;
	cout << "Berechnete Anzahl Meta-Atome: " << temp << endl;
	meta_atoms.reserve(temp);
	for (unsigned k = 1; k <= m; ++k){
			append_k_sized_meta_atom_subset(k);
	}
	cout <<  meta_atoms.size() << " Variablen im neuen Problem" << endl;
}

void PmTransformedTask::init_atoms(){
	for (int i = 0; i < parent_num_variables; ++i) {
		atom_helper.push_back(atoms.size());
		for(int j = 0; j < parent->get_variable_domain_size(i); ++j){
			atoms.push_back(make_pair(i,j));
		}
	}
}

void PmTransformedTask::init_findmap(){
	for(int index = 0; index < get_num_variables(); ++index){
		AtomCollection * ac = &meta_atoms[index];
		findmap[ac] = index;
	}
}

void PmTransformedTask::init_operators() {
    for (OperatorProxy op : parentProxy.get_operators()) {
		add_new_operator(op, AtomCollection());
		for(auto meta_atom : meta_atoms){
		// da laut definition nur bis m - 1
			if(meta_atom.size() > m - 1) break;
			if (check_disjoint(op.get_effects(), meta_atom)){
				add_new_operator(op, meta_atom);
			}
		}
	}
	cout << operators.size() << " Operatoren im neuen Problem" << endl;
}

void PmTransformedTask::init_initial_state(){
	State initial_state = parentProxy.get_initial_state();
	AtomCollection atms;
	for( FactProxy fp : initial_state){
		atms.push_back(get_atom(fp.get_variable().get_id(), fp.get_value()));
	}
	for (auto x : meta_atoms){
		if (includes(atms.begin(), atms.end(), x.begin(), x.end())) {
			initial_state_values.push_back(1);
		} else {
			initial_state_values.push_back(0);
		}
	}
}

void PmTransformedTask::init_goals(){
	AtomCollection goal_atms;
	for(int i = 0; i < parent->get_num_goals(); ++i){
		pair<int, int> fact = parent->get_goal_fact(i);
		goal_atms.push_back(get_atom(fact.first, fact.second));
	}
	sort(goal_atms.begin(), goal_atms.end());
	for(unsigned atm = 0; atm < meta_atoms.size(); ++atm){
        const AtomCollection &meta = meta_atoms[atm];
        if (includes(goal_atms.begin(), goal_atms.end(),
                     meta.begin(), meta.end()))
			goals.push_back(atm);
	}
}

vector<int>* PmTransformedTask::index_subset_of_size_k(const AtomCollection &set, int subset_size,
		vector<int> *out){
	AtomCollection l;
	index_subset_of_size_k(set, subset_size, l, set.begin(), out);
	return out;
}

void PmTransformedTask::index_subset_of_size_k(const AtomCollection &set, int subset_size,
		AtomCollection &l, AtomCollection::const_iterator it, vector<int> *out){
	if(subset_size == 0){
	    out->push_back(findmap.at(&l));
	    return;
    }
	for(AtomCollection::const_iterator i = it; i != set.end();){
	    l.push_back(*i);
	    index_subset_of_size_k(set, subset_size - 1, l, ++i, out);
	    l.pop_back();
	}
}

vector<PmTransformedTask::AtomCollection> PmTransformedTask::subset_of_size_k(
		const AtomCollection &set, int subset_size){
	AtomCollection l;
	vector<AtomCollection> out;
	subset_of_size_k(set, subset_size, l, set.begin(), &out);
	return out;
}

void PmTransformedTask::subset_of_size_k(const AtomCollection &set, int subset_size,
		AtomCollection &l, AtomCollection::const_iterator it, vector<AtomCollection> *out){
    if(subset_size == 0){
    	out->push_back(l);
        return;
    }
    for(AtomCollection::const_iterator i = it; i != set.end();){
        l.push_back(*i);
        subset_of_size_k(set, subset_size - 1, l, ++i, out);
        l.pop_back();
    }
}

void PmTransformedTask::print_atom_collection(const AtomCollection &l, char end_char, bool new_line){
	char white_space = (new_line) ? '\n' : ' ';
    for (int val : l)
        cout << val << white_space;
    if(new_line) return;
    cout << end_char;
}

void PmTransformedTask::append_k_sized_meta_atom_subset(int k){
	AtomCollection l;
	for (unsigned i = 0; i < atoms.size(); ++i){
		l.push_back(i);
	}
	vector<AtomCollection> ac = subset_of_size_k(l, k);
	for(const auto &x : ac){
		add_new_meta_atom(x);
	}
}

int PmTransformedTask::get_num_variables() const{
	 return meta_atoms.size();
}

const string PmTransformedTask::get_variable_name(int var) const{
	return get_meta_atom_fact_name(var);
}

int PmTransformedTask::get_variable_domain_size(int /*var*/) const {
	return 2;
}

const string PmTransformedTask::get_fact_name(int var, int value) const {
	if (value == 0) return NEGATION_PREFIX + get_meta_atom_fact_name(var);
	return get_meta_atom_fact_name(var);
}

int PmTransformedTask::get_operator_cost(int index, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
	return operators[index]->get_cost();
}

const string& PmTransformedTask::get_operator_name(int index, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
	return operators[index]->get_name();
}

int PmTransformedTask::get_num_operators() const {
	return operators.size();
}

int PmTransformedTask::get_num_operator_preconditions(int index, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
	return operators[index]->get_preconditions().size();
}

pair<int, int> PmTransformedTask::get_operator_precondition(int op_index,
		int fact_index, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
    int precondition = operators[op_index]->get_preconditions()[fact_index];
    return make_pair(precondition, 1);
}

int PmTransformedTask::get_num_operator_effects(int op_index, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
	return operators[op_index]->get_effects().size();
}

int PmTransformedTask::get_num_operator_effect_conditions(int /*op_index*/,
		int /*eff_index*/, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
	return 0;
}

pair<int, int> PmTransformedTask::get_operator_effect_condition(
		int op_index, int /*eff_index*/, int /*cond_index*/, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
	exit_cause_effect_conditions(op_index);
	return make_pair(-1, -1);
}

pair<int, int> PmTransformedTask::get_operator_effect(int op_index,
		int eff_index, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
	int effect = operators[op_index]->get_effects()[eff_index];
	return make_pair(effect, 1);
}

const GlobalOperator* PmTransformedTask::get_global_operator(int index, bool is_axiom) const {
	if(is_axiom) exit_cause_axiom();
	return parentProxy.get_operators()[index].get_global_operator();
}

int PmTransformedTask::get_num_goals() const {
	return goals.size();
}

pair<int, int> PmTransformedTask::get_goal_fact(int index) const {
    return make_pair(goals[index], 1);
}

vector<int> PmTransformedTask::get_initial_state_values() const {
	return initial_state_values;
}

vector<int> PmTransformedTask::get_state_values(const GlobalState &global_state) const {
	vector<int> parent_values = parent->get_state_values(global_state);
	vector<int> values(meta_atoms.size());
	int atm;
	for (unsigned var = 0; var < parent_values.size(); ++var){
		for(int fact = 0; fact < parent->get_variable_domain_size(var); ++fact){
			atm = get_atom(var, fact);
			values[atm] = (fact == parent_values[var]) ? 1 : 0;
		}
	}
	for (unsigned meta_atm = parent_values.size(); meta_atm < meta_atoms.size(); ++meta_atm){
		int val = 1;
		for (int atom : meta_atoms[meta_atm]){
			if (values[atom] == 0) val = 0;
		}
		values[meta_atm] = val;
	}
    return values;
}

void PmTransformedTask::add_new_meta_atom(const AtomCollection &atoms){
	meta_atoms.push_back(atoms);
}

void PmTransformedTask::add_new_operator(OperatorProxy &orig_op, const AtomCollection &meta_atom) {
	int cost = orig_op.get_cost();
	string name = orig_op.get_name();
	GeneratedOperator *op = new GeneratedOperator(name, make_preconditions(orig_op, meta_atom),
			make_effects(orig_op, meta_atom), cost);
	operators.push_back(op);
	parent_operator.push_back(orig_op.get_id());
}

vector<int> PmTransformedTask::make_preconditions(const OperatorProxy &orig_op, const AtomCollection &meta_atom){
	AtomCollection atms;
	vector<int> pre;
    for (FactProxy fp : orig_op.get_preconditions())
        atms.push_back(get_atom(fp.get_variable().get_id(), fp.get_value()));
 	for (int atom : meta_atom){
 		atms.push_back(atom);
 	}
 	atms = remove_doubles(atms);
 	for (int k = 1; unsigned(k) <= m; k++){
 		pre = *index_subset_of_size_k(atms, k, &pre);
 	}
	return pre;
}

// TODO prüfen ob nicht effizienter wenn man dies wie make_preconditions implementiert,
// ... also auf (c vereinigt add(a))!= {} verzichtet
vector<int> PmTransformedTask::make_effects(const OperatorProxy &orig_op, const AtomCollection &meta_atom){
	AtomCollection op_eff;
 	AtomCollection atms;
    for (EffectProxy ep : orig_op.get_effects()) {
        FactProxy fp = ep.get_fact();
        int eff = get_atom(fp.get_variable().get_id(), fp.get_value());
 		op_eff.push_back(eff);
 		atms.push_back(eff);
 	}
 	for (int atm : meta_atom){
 		atms.push_back(atm);
 	}
 	vector<int> effects;
 	atms = remove_doubles(atms);
 	sort(op_eff.begin(), op_eff.end());
 	for (int k = 1; unsigned(k) <= m; ++k){
 		vector<AtomCollection>  ac = subset_of_size_k(atms, k); // TODO direkte Methode
 		for(AtomCollection &x : ac){
 			if(contains_at_least_one_list_element(x, op_eff)){
 				int meta_atm = findmap.at(&x);
 				effects.push_back(meta_atm);
 			}
 		}
 	}
 	return effects;
}

int PmTransformedTask::get_atom(int var_index, int fact_index) const{
	return atom_helper[var_index] + fact_index;
}

string PmTransformedTask::get_atom_fact_name(int atom, int fact) const{
    int var = atoms[atom].first, val = atoms[atom].second;
	if(fact == 0) return NEGATION_PREFIX + parent->get_fact_name(var, val);
	return parent->get_fact_name(var, val);
}

string PmTransformedTask::get_meta_atom_fact_name(int meta_atom) const{
	string s;
	for (auto atom : meta_atoms[meta_atom]){
		s += get_atom_fact_name(atom);
		if(atom == meta_atoms[meta_atom].back()) return s;
		s += META_ATOM_JOINER;
	}
	cerr << "!!!\n!!!\n!!!\n!!!\n!!!\nVerdammt, da ging was falsch\n!!!\n!!!\n!!!\n!!!\n!!!\n " << endl;
	return s;
}

bool PmTransformedTask::check_disjoint(const EffectsProxy &effects, const AtomCollection &meta_atom){
	for (int atom : meta_atom){
        for (EffectProxy eff : effects) {
            FactProxy fact = eff.get_fact();
            if (atoms[atom].first == fact.get_variable().get_id())
                return false;
		}
	}
	return true;
}

// return true, if which contains at least one element of what. Both lists have to be sorted
bool PmTransformedTask::contains_at_least_one_list_element(const AtomCollection &which, const AtomCollection &what){
	auto it = which.begin();
	auto jt = what.begin();
	while (it != which.end() && jt != what.end()){
		if(*it == *jt) return true;
		if(*it > *jt) ++jt;
		else ++it;
	}
	return false;
}

PmTransformedTask::AtomCollection PmTransformedTask::remove_doubles(const AtomCollection &in){
	AtomCollection out = in;
	sort(out.begin(), out.end());
	auto it = unique(out.begin(), out.end());
	out.resize(distance(out.begin(), it));
	return out;
}

PmTransformedTask::AtomCollection PmTransformedTask::get_merged(const AtomCollection &lhs, const AtomCollection &rhs){
	AtomCollection out;
	auto it = lhs.begin();
	auto jt = rhs.begin();
	while (it != lhs.end() && jt != rhs.end()){
		if(*it == *jt){
			if(out.back() != *it)
				out.push_back(*it);
			++it;
			++jt;
		}else if(*it > *jt){
			if(out.back() != *jt)
				out.push_back(*jt);
			++jt;
		}else{
			if(out.back() != *it)
				out.push_back(*it);
			++it;
		}
	}
	return out;
}

int PmTransformedTask::bin(int n, int k){
	if (k==0) return 1;
	if (2*k > n) return bin(n, n-k);
	int temp = n - k + 1;
	for (int i = 2; i<=k; ++i){
		temp*=(n-k+i);
		temp/=i;
	}
	return temp;
}

void PmTransformedTask::exit_cause_axiom() const{
	 cerr << "This configuration does not support axioms!"
	             << endl << "Terminating." << endl;
	exit_with(EXIT_UNSUPPORTED);
}

void PmTransformedTask::exit_cause_effect_conditions(int op_index) const{
    cerr << "This configuration does not support conditional effects "
         << "(operator " << operators[op_index]->get_name() << ")!" << endl
         << "Terminating." << endl;
    exit_with(EXIT_UNSUPPORTED);
}

static shared_ptr<AbstractTask> _parse(OptionParser &parser) {
	parser.add_option<int>("m", "Subset size", "2", Bounds("1", "infinity"));
	parser.add_option<shared_ptr<AbstractTask>>("transform",
			"Parent task transformation", "no_transform");
	add_cost_type_option_to_parser(parser);
	Options opts = parser.parse();
    parser.document_language_support("conditional effects", "not supported");
    parser.document_language_support("axioms", "not supported");
	if (parser.dry_run())
		return nullptr;
	else
		return make_shared<PmTransformedTask>(opts);
}

static PluginShared<AbstractTask> _plugin("pm_transform", _parse);
